package filegateway

import (
	"fmt"
	"io"
	"path"

	nrpc "github.com/nats-rpc/gogo-nrpc"
	stdlib "orus.io/xbus/xbus-stdlib/lib"
	xbus "xbus.io/go-xbus/v3"
)

/*
```yaml
actor:
  kind: consumer
  service:
    type: file-gateway
    settings:
       uri: scp://remote/directory/
       secrets: path/to/passwords-file.yaml
       poll-delay: 10s
       wait-for: deletion
```
*/

var defaultConsumerLog xbus.Logger

// Consumer save the received files to a filesystem
type Consumer struct {
	fs        Filesystem
	targetDir string
	log       xbus.Logger

	temporaryErrors bool
}

// NewConsumer creates a Consumer
func NewConsumer(actor *xbus.Actor) xbus.ActorService {
	options, err := LoadFilesystemOptions(actor, "target")
	if err != nil {
		actor.Log.Errorf("Could not load 'target' setting: %s", err)
		return nil
	}
	fs, err := GetFilesystem(options.Scheme, options)
	if err != nil {
		if actor.Log != nil {
			actor.Log.Errorf("Could not get filesystem: %s", err)
		} else {
			fmt.Println("Could not get filesystem:", err)
		}

		return nil
	}
	log := defaultConsumerLog
	if actor.Client != nil {
		log = actor.Client.Log
	}
	return &Consumer{
		fs:              fs,
		targetDir:       options.Path,
		temporaryErrors: actor.GetBoolSettingD("temporary-errors", false),
		log:             log,
	}
}

func (c *Consumer) fsError(op string, targetPath string, err error) error {
	if c.temporaryErrors {
		return &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[filegateway] Error %s file '%s': %s.", op, targetPath, err),
		}
	}
	return err
}

// Process ...
func (c *Consumer) Process(apc *xbus.ActorProcessingContext) error {
	reader, err := apc.ReadEnvelope("default")
	if err != nil {
		return err
	}
	eventTypes := reader.EventTypes()
	if len(eventTypes) != 1 {
		return fmt.Errorf("Can only handle envelopes with 1 message, not %d", len(eventTypes))
	}
	fReader := stdlib.NewFileMessageReader(reader.GetItemChan(eventTypes[0]))
	meta, err := fReader.Metadata()
	if err != nil {
		return err
	}
	body, err := fReader.Body()
	if err != nil {
		return err
	}
	defer func() {
		if err := body.Close(); err != nil {
			c.log.Errorf("filegateway.Consumer: error closing the file body: %s", err)
		}
	}()
	targetPath := path.Join(c.targetDir, meta.Filename)
	c.log.Debugf("Writing %s to %s", targetPath, c.fs)
	writer, err := c.fs.Write(targetPath, -1)
	if err != nil {
		return c.fsError("opening", targetPath, err)
	}
	if _, err = io.Copy(writer, body); err != nil {
		if err := writer.Close(); err != nil {
			xbus.GetZeroLogger(c.log).Warn().Err(err).Msg("error closing the writer during an error handling")
		}
		return c.fsError("writing", targetPath, err)
	}
	if err := writer.Close(); err != nil {
		return c.fsError("closing", targetPath, err)
	}
	c.log.Noticef("Successfully written %s to %s", targetPath, c.fs)
	return nil
}

func init() {
	xbus.RegisterActorService("file-gateway-consumer", NewConsumer)
}
