package filegateway

import (
	"context"
	"testing"
	"time"

	nrpc "github.com/nats-rpc/gogo-nrpc"
	"github.com/stretchr/testify/assert"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/envelope"
	"xbus.io/go-xbus/v3/fullenv"
	"xbus.io/go-xbus/v3/testutils"
)

func setConsumerLogger(tb testing.TB) func() {
	savedLogger := defaultConsumerLog
	defaultConsumerLog = testutils.GetTestLogger(tb)
	return func() {
		defaultConsumerLog = savedLogger
	}
}

func TestConsumer(t *testing.T) {
	t.Run("basics", func(t *testing.T) {
		defer setConsumerLogger(t)()
		tester := testutils.NewActorServiceTester(t, "file-service-name", NewConsumer)
		tester.SetSettings(map[string]interface{}{
			"target": "mock://testconsumer/targetdir/",
		})
		envelope := xbus.ForgeEnvelopeFromItems("test", []api.Item{
			api.Item(`{"filename":"file.txt"}`),
			api.Item("Hello World!"),
		})
		tester.SetInputEnvelope("default", &envelope)

		state, err := tester.RunProcess()
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, api.ActorProcessingState_SUCCESS, state.Status)
		assert.Equal(t, "Hello World!", string(getMemoryFilestem("testconsumer").Files["/targetdir/file.txt"]))
	})
	t.Run("temporary errors returns SERVERTOOBUSY", func(t *testing.T) {
		defer setConsumerLogger(t)()
		tester := testutils.NewActorServiceTester(t, "file-service-name", NewConsumer)
		tester.SetSettings(map[string]interface{}{
			"target":           "mock://testconsumer/targetdir/",
			"target.fserror":   "onwrite",
			"temporary-errors": true,
		})
		envelope := xbus.ForgeEnvelopeFromItems("test", []api.Item{
			api.Item(`{"filename":"file.txt"}`),
			api.Item("Hello World!"),
		})
		tester.SetInputEnvelope("default", &envelope)

		_, err := tester.RunProcess()
		if nErr, ok := err.(*nrpc.Error); assert.True(t, ok, "Error is: %s", err) {
			assert.Equal(t, nrpc.Error_SERVERTOOBUSY, nErr.Type)
			assert.Equal(t, "[filegateway] Error opening file '/targetdir/file.txt': write error.", nErr.Message)
		}
	})
	t.Run("fullenv temporary errors", func(t *testing.T) {
		defer setConsumerLogger(t)()
		env := fullenv.NewFullenvForTest(t, "filegateway", "consumer_test.yaml")
		if err := env.Start(); err != nil {
			t.Fatal(err)
		}
		defer env.Stop()
		client, err := env.LoadClient("client")
		if err != nil {
			t.Fatal(err)
		}
		if err := client.Startup(true); err != nil {
			t.Fatal(err)
		}
		defer client.Shutdown()

		emitter := client.GetActorByName("emitter")
		if emitter == nil {
			t.Fatal("No 'emitter'")
		}
		msg, w, err := envelope.NewMessageChunkWriter("msg")
		if err != nil {
			t.Fatal(err)
		}
		if !assert.NoError(t, w.WriteString(`{"filename": "test.txt"}`)) ||
			!assert.NoError(t, w.WriteString(`SomeContent`)) ||
			!assert.NoError(t, w.Close()) {
			t.FailNow()
		}
		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
		defer cancel()
		emit, err := emitter.Emit(ctx, envelope.MustNewEnvelope(msg))
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		completed := make(chan error)
		go func() {
			err := emit.WaitCompletion()
			t.Log("WaitCompletion is over, result is", err)
			completed <- err
		}()
		// Now the process should remain 'RUNNING' and never end because the
		// consumer returns SERVERTOOBUSY
		select {
		case <-ctx.Done():
			// timeout reached, as expected
			t.Log("Timeout reached as expected")
		case err := <-completed:
			if ctx.Err() != nil && err == ctx.Err() {
				t.Log("Timeout reached as expected (indirect)")
			} else {
				t.Error("The process terminated with", err)
			}
		}
	})
}
