// Copyright 2019 Orus. All rights reserved.

package filegateway

import (
	"context"
	"fmt"
	"io"
	"net/url"
	"strings"
	"time"

	xbus "xbus.io/go-xbus/v3"
)

// CustomOptions ...
type CustomOptions map[string]interface{}

// FilesystemOptions ...
type FilesystemOptions struct {
	Name            string
	Scheme          string
	Host            string
	Port            string
	Username        string
	Password        string
	HasPassword     bool
	Path            string
	PollingInterval time.Duration
	Custom          CustomOptions
}

func (fo FilesystemOptions) String() string {
	url := fo.Scheme + "://" + fo.Username + "@" + fo.Host + "/" + fo.Path
	query := []string{}
	for key, value := range fo.Custom {
		svalue := fmt.Sprintf("%v", value)
		if len(svalue) > 13 {
			svalue = svalue[:10] + "..."
		}
		query = append(query, key+"="+svalue)
	}
	if len(query) != 0 {
		url += "?" + strings.Join(query, "&")
	}
	return url
}

// CustomOption returns an option, if present
func (fo FilesystemOptions) CustomOption(key string) (interface{}, bool) {
	value, ok := fo.Custom[key]
	return value, ok
}

// GetCustomStringD returns a string option with a default value
func (fo FilesystemOptions) GetCustomStringD(key string, defaultValue string) (value string, err error) {
	err = fo.CustomStringD(key, &value, defaultValue)
	return
}

// CustomStringD reads a string option, if present, or defaultValue if not
func (fo FilesystemOptions) CustomStringD(key string, target *string, defaultValue string) error {
	if value, ok := fo.CustomOption(key); ok {
		if s, ok := value.(string); ok {
			*target = s
			return nil
		}
		return fmt.Errorf("Option '%s' must be a string. Got: %T", key, value)
	}
	*target = defaultValue
	return nil
}

// CustomIntD reads a int option if present, or defaultValue if not
func (fo FilesystemOptions) CustomIntD(key string, target *int, defaultValue int) error {
	if value, ok := fo.CustomOption(key); ok {
		if s, ok := value.(int); ok {
			*target = s
			return nil
		}
		return fmt.Errorf("Option '%s' must be a int. Got: %T", key, value)
	}
	*target = defaultValue
	return nil
}

// NewFilesystemOptions initializes a FilesystemOptions from a uri
func NewFilesystemOptions(sURI string) (FilesystemOptions, error) {
	uri, err := url.Parse(sURI)
	if err != nil {
		return FilesystemOptions{}, err
	}
	options := FilesystemOptions{
		Scheme: uri.Scheme,
		Host:   uri.Hostname(),
		Port:   uri.Port(),
		Path:   uri.Path,
		Custom: make(CustomOptions),
	}
	if user := uri.User; user != nil {
		options.Username = user.Username()
		options.Password, options.HasPassword = user.Password()
	}
	value, err := options.GetCustomStringD("polling-interval", "10s")
	if err != nil {
		return options, err
	}
	options.PollingInterval, err = time.ParseDuration(value)
	if err != nil {
		return options, fmt.Errorf("could not parse 'polling-interval': %s", err)
	}
	delete(options.Custom, "polling-interval")
	return options, nil
}

// LoadFilesystemOptions loads a filesytem options from a actor service settings
func LoadFilesystemOptions(actor *xbus.Actor, name string) (FilesystemOptions, error) {
	options, err := NewFilesystemOptions(actor.GetStringSetting(name))
	if err != nil {
		return options, err
	}
	options.Name = name
	prefix := name + "."
	for key, value := range actor.Options.Service.Settings {
		if strings.HasPrefix(key, prefix) {
			options.Custom[key[len(prefix):]] = value
		}
	}
	return options, nil
}

// FilesystemFactory is a function that instanciate a Filesystem
type FilesystemFactory func(FilesystemOptions) (Filesystem, error)

// Filesystem ...
type Filesystem interface {
	String() string
	Write(path string, size int) (io.WriteCloser, error)
	Observe(ctx context.Context, pattern string, logger xbus.Logger) (<-chan string, error)
	Read(path string) (io.ReadCloser, error)
	WaitForDeletion(ctx context.Context, path string) error
	Rename(path, newpath string) error
	Delete(path string) error
}

// GetFilesystem instanciate a Filesytem for a given scheme and options
func GetFilesystem(scheme string, options FilesystemOptions) (Filesystem, error) {
	if fsFactory, ok := fsFactories[scheme]; ok {
		fs, err := fsFactory(options)
		if fs == nil && err == nil {
			err = fmt.Errorf("%s factory return a nil Filesystem with no err msg", scheme)
		}
		return fs, err
	}
	return nil, fmt.Errorf("Invalid scheme: %s", scheme)
}

var fsFactories = make(map[string]FilesystemFactory)

// RegisterFilesystem registers a new filesystem type
func RegisterFilesystem(scheme string, factory FilesystemFactory) error {
	fsFactories[scheme] = factory
	return nil
}

// MustRegisterFilesystem registers a filesystem or panics if it fails
func MustRegisterFilesystem(scheme string, factory FilesystemFactory) {
	if err := RegisterFilesystem(scheme, factory); err != nil {
		panic(err)
	}
}
