package routing

import (
	"database/sql"
	"errors"
	"fmt"

	"github.com/Masterminds/squirrel"
	"github.com/jmoiron/sqlx"
	"github.com/lann/builder"

	// load the postgresql driver
	_ "github.com/lib/pq"

	"orus.io/xbus/xbus-stdlib/service"
)

func setPlaceHolderFormat(query squirrel.Sqlizer) squirrel.Sqlizer {
	return builder.Set(query, "PlaceholderFormat", squirrel.Dollar).(squirrel.Sqlizer)
}

// NewPostgresTableStore returns a postgresql based TableStore implementation
func NewPostgresTableStore(dsn, tablename, keycolumn, valuecolumn string, autocreate bool) (*PostgresTableStore, error) {
	db, err := sqlx.Open("postgres", dsn)
	if err != nil {
		return nil, err
	}

	// Check that the table exists
	var regclass sql.NullString
	if err := db.Get(&regclass, "SELECT to_regclass($1)", tablename); err != nil {
		_ = db.Close()
		return nil, err
	}
	if !regclass.Valid {
		if !autocreate {
			return nil, fmt.Errorf("table %s does not exist (try enabling 'autocreate')", tablename)
		}
		query := fmt.Sprintf(
			"CREATE TABLE %s (%s VARCHAR PRIMARY KEY, %s VARCHAR)",
			tablename, keycolumn, valuecolumn)
		if _, err := db.Exec(query); err != nil {
			_ = db.Close()
			return nil, err
		}
	}

	// check that the columns exist
	rows, err := db.Query(fmt.Sprintf(
		"SELECT %[2]s, %[3]s FROM %[1]s LIMIT 0", tablename, keycolumn, valuecolumn))
	if err != nil {
		_ = db.Close()
		return nil, err
	}
	_ = rows.Close()

	setStmt, err := db.Preparex(fmt.Sprintf(
		"INSERT INTO %[1]s (%[2]s, %[3]s) VALUES ($1, $2) ON CONFLICT (%[2]s) DO UPDATE SET %[3]s=$2",
		tablename, keycolumn, valuecolumn,
	))
	if err != nil {
		_ = db.Close()
		return nil, err
	}
	getStmt, err := db.Preparex(fmt.Sprintf(
		"SELECT %[3]s FROM %[1]s WHERE %[2]s = $1",
		tablename, keycolumn, valuecolumn,
	))
	if err != nil {
		_ = db.Close()
		return nil, err
	}
	deleteStmt, err := db.Preparex(fmt.Sprintf(
		"DELETE FROM %[1]s WHERE %[2]s = $1",
		tablename, keycolumn,
	))
	if err != nil {
		_ = db.Close()
		return nil, err
	}
	return &PostgresTableStore{
		db:         db,
		setStmt:    setStmt,
		getStmt:    getStmt,
		deleteStmt: deleteStmt,
	}, nil
}

// PostgresTableStore ...
type PostgresTableStore struct {
	db         *sqlx.DB
	setStmt    *sqlx.Stmt
	getStmt    *sqlx.Stmt
	deleteStmt *sqlx.Stmt
}

// Set a value
func (s *PostgresTableStore) Set(key, value string) error {
	if _, err := s.setStmt.Exec(key, value); err != nil {
		return err
	}
	return nil
}

// Get a value
func (s *PostgresTableStore) Get(key string) (string, error) {
	var value string
	if err := s.getStmt.Get(&value, key); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return "", ErrNoSuchKey
		}
		return "", err
	}
	return value, nil
}

// Delete a value
func (s *PostgresTableStore) Delete(key string) error {
	if _, err := s.deleteStmt.Exec(key); err != nil {
		return err
	}
	return nil
}

// Close ...
func (s *PostgresTableStore) Close() error {
	return s.db.Close()
}

func postgresTableStoreFactory(prefix string, settings service.Settings) (TableStore, error) {
	dsn, err := settings.String(prefix + "dsn")
	if err != nil {
		return nil, err
	}
	tablename, err := settings.String(prefix + "tablename")
	if err != nil {
		return nil, err
	}
	keycolumn, err := settings.StringD(prefix+"keycolumn", "key")
	if err != nil {
		return nil, err
	}
	valuecolumn, err := settings.StringD(prefix+"valuecolumn", "value")
	if err != nil {
		return nil, err
	}
	autocreate, err := settings.BoolD(prefix+"autocreate", false)
	if err != nil {
		return nil, err
	}
	return NewPostgresTableStore(dsn, tablename, keycolumn, valuecolumn, autocreate)
}

func init() {
	MustRegisterTableStore("postgres", postgresTableStoreFactory)
}
