package service

import (
	"context"
	"io"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/envelope"
	"xbus.io/go-xbus/v3/fullenv"
)

func TestService(t *testing.T) {
	t.Run("worker-forward", func(t *testing.T) {
		workerResult := make(chan error, 1)
		RegisterWorkerFunc("TestService-worker", func(
			ctx context.Context, req api.ActorProcessRequest,
			inputs Inputs, outputs Outputs, log xbus.ZeroLogger,
		) error {
			e, err := inputs.Get("default")
			if err != nil {
				workerResult <- err
				return err
			}
			err = outputs.Forward("default", e.ID())
			workerResult <- err
			return err
		})
		defer UnregisterWorker("TestService-worker")

		consumerResult := make(chan error, 1)
		RegisterConsumerFunc("TestService-consumer", func(
			ctx context.Context, req api.ActorProcessRequest,
			inputs Inputs, log xbus.ZeroLogger,
		) error {
			_, err := inputs.Get("default")
			consumerResult <- err
			return err
		})
		defer UnregisterConsumer("TestService-consumer")

		env := fullenv.NewFullenvForTest(t, "service", "service_test.yaml")
		require.NoError(t, env.Start())
		defer env.Stop()
		client, err := env.LoadClient("client")
		require.NoError(t, err)
		require.NoError(t, client.Startup(true))
		defer client.Shutdown()

		emitter := client.GetActorByName("emitter")
		if emitter == nil {
			t.Fatal("No 'emitter'")
		}

		ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
		defer cancel()

		emission, err := emitter.Emit(
			ctx, envelope.MustNewEnvelope(envelope.MustNewTextMessage("msg", "body")))
		require.NoError(t, err)
		require.NoError(t, emission.WaitCompletion())

		if err, ok := <-workerResult; assert.True(t, ok) {
			assert.NoError(t, err)
		}
		if err, ok := <-consumerResult; assert.True(t, ok) {
			assert.NoError(t, err)
		}
	})
	t.Run("big-envelope", func(t *testing.T) {
		workerResult := make(chan error, 1)
		RegisterWorkerFunc("TestService-worker", func(
			ctx context.Context, req api.ActorProcessRequest,
			inputs Inputs, outputs Outputs, log xbus.ZeroLogger,
		) error {
			e, err := inputs.Get("default")
			if err != nil {
				workerResult <- err
				return err
			}
			out, err := envelope.NewOpenedEnvelope()
			if err != nil {
				workerResult <- err
				return err
			}
			go func() {
				err := envelope.ForEachMessage(e, func(msg envelope.Message) error {
					t.Log("Reading message", msg.MsgType())
					outMsg, writer, err := envelope.NewMessageWriter(msg.MsgType())
					if err != nil {
						return err
					}
					if err := out.Add(outMsg); err != nil {
						return err
					}
					t.Log("Copying message", msg.MsgType())
					if _, err := io.Copy(writer, msg.Reader()); err != nil {
						return err
					}
					t.Log("Closing message", msg.MsgType())
					if err := writer.Close(); err != nil {
						return err
					}
					return nil
				})
				if err != nil {
					out.CloseWithError(err)
				} else {
					assert.NoError(t, out.Close())
				}
			}()

			err = outputs.Send("default", out)
			workerResult <- err
			return err
		})
		defer UnregisterWorker("TestService-worker")

		consumerResult := make(chan error, 1)
		RegisterConsumerFunc("TestService-consumer", func(
			ctx context.Context, req api.ActorProcessRequest,
			inputs Inputs, log xbus.ZeroLogger,
		) error {
			_, err := inputs.Get("default")
			consumerResult <- err
			return err
		})
		defer UnregisterConsumer(t.Name() + "TestService-consumer")

		env := fullenv.NewFullenvForTest(t, "service", "service_test.yaml")
		require.NoError(t, env.Start())
		defer env.Stop()
		client, err := env.LoadClient("client")
		require.NoError(t, err)
		require.NoError(t, client.Startup(true))
		defer client.Shutdown()

		emitter := client.GetActorByName("emitter")
		if emitter == nil {
			t.Fatal("No 'emitter'")
		}

		ctx, cancel := context.WithTimeout(context.Background(), 40*time.Second)
		defer cancel()

		msg, msgW, err := envelope.NewMessageWriter("msg")
		require.NoError(t, err)
		go func() {
			for i := 0; i < 1024*1024; i++ {
				_, err := msgW.Write([]byte("azertyuiopqsdfghjklmwxcvbn0123456789"))
				assert.NoError(t, err)
			}
			assert.NoError(t, msgW.Close())
			t.Log("Done writing")
		}()
		emission, err := emitter.Emit(
			ctx, envelope.MustNewEnvelope(msg))
		require.NoError(t, err)
		require.NoError(t, emission.WaitCompletion())

		if err, ok := <-workerResult; assert.True(t, ok) {
			assert.NoError(t, err)
		}
		if err, ok := <-consumerResult; assert.True(t, ok) {
			assert.NoError(t, err)
		}
	})
}
