package svctest

import (
	"context"
	"testing"

	"github.com/stretchr/testify/require"
	"orus.io/xbus/xbus-stdlib/service"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/testutils"
)

// ConsumerTester helps testing a Consumer
type ConsumerTester struct {
	t        *testing.T
	factory  service.ConsumerFactory
	actor    *api.Actor
	settings service.Settings
}

// Run calls a consumer implementation with the given parameters
func (tester ConsumerTester) Run(inputs Inputs) error {
	return tester.RunContext(context.Background(), inputs)
}

// RunContext calls a consumer implementation with the given parameters, and
// returns the output envelopes
func (tester ConsumerTester) RunContext(ctx context.Context, inputs Inputs) error {
	log := testutils.GetTestLogger(tester.t)
	consumer, err := tester.factory(tester.actor, tester.settings)
	if err != nil {
		return err
	}
	if i, ok := consumer.(service.WithStartup); ok {
		require.NoError(tester.t, i.Startup())
	}
	requestInputs := make([]*api.ActorProcessRequest_Input, 0, len(inputs))
	for name, input := range inputs {
		var (
			envelope *api.Envelope
			position *api.EnvelopePosition
		)
		if input != nil {
			envelope = &api.Envelope{
				ID: input.ID(),
			}
			p := api.EnvelopePositionStart(input.ID())
			position = &p
		}

		requestInput := &api.ActorProcessRequest_Input{
			Name:     name,
			Close:    input == nil,
			Envelope: envelope,
			Position: position,
		}

		requestInputs = append(requestInputs, requestInput)
	}
	err = consumer.Process(ctx,
		api.ActorProcessRequest{
			Context: &api.ProcessingContext{
				ProcessID: DefaultProcessID,
				NodeID:    DefaultNodeID,
			},
			Inputs: requestInputs,
		},
		inputs, log.ZeroLogger)
	if i, ok := consumer.(service.WithShutdown); ok {
		require.NoError(tester.t, i.Shutdown())
	}
	return err
}

// TestConsumer returns a ConsumerTester for a given ConsumerFactory
func TestConsumer(t *testing.T,
	factory service.ConsumerFactory,
	actor *api.Actor, settings service.Settings) *ConsumerTester {
	return &ConsumerTester{
		t: t, factory: factory, actor: actor, settings: settings,
	}
}

// TestConsumerFunc returns a ConsumerTester for a given ConsumerFunc
func TestConsumerFunc(t *testing.T, fn service.ConsumerFunc) *ConsumerTester {
	return &ConsumerTester{
		t: t, actor: nil, settings: nil,
	}
}
