package service

import (
	"context"
	"fmt"
	"runtime/debug"

	"xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
)

func catchErrors(log xbus.Logger) {
	if r := recover(); r != nil {
		xbus.GetZeroLogger(log).Error().Msgf("Caught error: %s\n%s", r, debug.Stack())
	}
}

// WorkerFunc is a function implementing a consumer
type WorkerFunc func(context.Context, api.ActorProcessRequest, Inputs, Outputs, xbus.ZeroLogger) error

// Worker ...
type Worker interface {
	Process(context.Context, api.ActorProcessRequest, Inputs, Outputs, xbus.ZeroLogger) error
}

// WorkerFactory is a factory of Worker
type WorkerFactory func(*api.Actor, Settings) (Worker, error)

// RegisterWorker registers a consumer
func RegisterWorker(name string, workerFactory WorkerFactory) (err error) {
	defer func() {
		if r := recover(); r != nil {
			err = fmt.Errorf("%v", r)
		}
	}()
	xbus.RegisterActorService(name, newWorkerServiceFactory(workerFactory))
	return
}

// RegisterWorkerFunc registers a consumer
func RegisterWorkerFunc(name string, process WorkerFunc) error {
	return RegisterWorker(name, func(*api.Actor, Settings) (Worker, error) { return process, nil })
}

// UnregisterWorker unregisters a worker
func UnregisterWorker(name string) {
	xbus.UnregisterActorService(name)
}

// Process calls the function
func (fn WorkerFunc) Process(ctx context.Context, req api.ActorProcessRequest, inputs Inputs, outputs Outputs, log xbus.ZeroLogger) error {
	return fn(ctx, req, inputs, outputs, log)
}

func newWorkerServiceFactory(factory WorkerFactory) func(actor *xbus.Actor) xbus.ActorService {
	return func(actor *xbus.Actor) xbus.ActorService {
		apiActor := actor.Options.Actor
		settings := Settings(actor.Options.Service.Settings)
		worker, err := factory(&apiActor, settings)
		if err != nil {
			panic(err)
		}
		return &actorService{
			actor:   actor,
			service: worker,
			process: workerProcess(worker),
		}
	}
}

func workerProcess(worker Worker) func(
	ctx context.Context, svc *actorService, req api.ActorProcessRequest,
	send func(api.ActorProcessingState),
) error {
	return func(
		ctx context.Context, svc *actorService, req api.ActorProcessRequest,
		send func(api.ActorProcessingState),
	) error {
		log := svc.actor.Client.Log.(xbus.ZeroLogger)
		inputs := realInputs{
			ctx:       ctx,
			client:    svc.actor.Client,
			actorID:   svc.actor.ID,
			actorKind: "worker",
			inputs:    req.Inputs,
		}
		outputs := realOutputs{
			ctx:               ctx,
			actor:             svc.actor,
			processingContext: *req.Context,
		}
		return worker.Process(ctx, req, inputs, outputs, log)
	}
}
